import pytest

import jsoml

import os
from pathlib import Path


TESTS_DIR = Path(__file__).parent
WEBSTRACT_DIR = TESTS_DIR / "cases/webstract"


@pytest.mark.parametrize("case", os.listdir(WEBSTRACT_DIR))
def test_webstract_xml_dumps_n_loads(case):
    expected = jsoml.load(WEBSTRACT_DIR / case / "output.xml")
    got = jsoml.loads(jsoml.dumps(expected))
    assert got == expected



from . import webstract

# standard library
import os
from pathlib import Path


class DocLoader:
    def __init__(self, cache, eprinter_config=None):
        self.cache = Path(cache)
        self.pandoc_opts = eprinter_config.pandoc_opts

    def webstract_from_edition(self, edition):
        work_path = self.cache / "arc" / str(edition.dsi)
        cached = self.cache / "epijats" / str(edition.dsi) / "webstract.xml"
        if cached.exists():
            ret = webstract.Webstract.load_xml(cached)
            ret.source.path = work_path
        else:
            if not work_path.exists():
                edition.work_copy(work_path)
            if work_path.is_dir():
                from . import jats

                ret = jats.webstract_from_jats(work_path, self.pandoc_opts)
            else:
                raise ValueError(f"Unknown digital object type at {edition.dsi}")

            edidata = dict(edid=str(edition.edid), base_dsi=str(edition.suc.dsi))
            latest_edid = edition.suc.latest(edition.unlisted).edid
            if latest_edid > edition.edid:
                edidata["newer_edid"] = str(latest_edid)
            ret['edition'] = edidata

            os.makedirs(cached.parent, exist_ok=True)
            ret.dump_xml(cached)

        return ret



import os, shutil, tempfile


def up_to_date(target, source):
    last_update = os.path.getmtime(target) if target.exists() else 0
    return last_update > os.path.getmtime(source)


def copytree_nostat(src, dst):
    """like shutil but avoids calling copystat so SELinux context is not copied"""

    os.makedirs(dst, exist_ok=True)
    for srcentry in os.scandir(src):
        dstentry = os.path.join(dst, srcentry.name)
        if srcentry.is_dir():
            copytree_nostat(srcentry, dstentry)
        else:
            shutil.copy(srcentry, dstentry)
    return dst


def swhid_from_files(path):
    import git

    if not path.is_dir():
        return "swh:1:cnt:" + git.Git().hash_object(path)
    with tempfile.TemporaryDirectory() as tmpdir:
        repo = git.Repo.init(tmpdir)
        g = git.Git(path)  # path is working dir
        g.set_persistent_git_options(git_dir=repo.git_dir)
        g.add(".")
        return "swh:1:dir:" + g.write_tree()



from epijats import Eprint, EprinterConfig, Webstract
from epijats.util import copytree_nostat

# std lib
import argparse, importlib, logging, os, shutil, subprocess, sys, tempfile
from pathlib import Path


def enable_weasyprint_logging():
    from weasyprint import LOGGER

    LOGGER.setLevel(logging.INFO)
    LOGGER.addHandler(logging.StreamHandler())


class Main:
    def __init__(self, cmd_line_args=None):
        self.parser = argparse.ArgumentParser(description="Eprint JATS")
        self.parser.add_argument("inpath", type=Path, help="input directory/path")
        self.parser.add_argument("outpath", type=Path, help="output directory/path")
        self.parser.add_argument(
            "--from",
            dest="inform",
            choices=["jats", "json", "yaml", "jsoml", "html"],
            default="jats",
            help="format of source",
        )
        self.parser.add_argument(
            "--to",
            dest="outform",
            choices=["json", "yaml", "jsoml", "html", "html+pdf", "pdf"],
            default="pdf",
            help="format of target",
        )
        self.parser.add_argument(
            "--no-web-fonts",
            default=False,
            action="store_true",
            help="Do not use online web fonts",
        )
        self.parser.parse_args(cmd_line_args, self)

        self.config = EprinterConfig(dsi_base_url="https://perm.pub")
        self.config.embed_web_fonts = not self.no_web_fonts

    def run(self):
        self.check_conversion_order()
        if self.just_copy():
            return

        webstract = self.load_webstract()
        if webstract is None:
            assert self.inform == "html" and self.outform == "pdf"
            Eprint.html_to_pdf(self.inpath, self.outpath)
            return
        self.convert(webstract)

    def check_conversion_order(self):
        format_stages = {
            'jats': 0,
            'json': 1,
            'yaml': 1,
            'jsoml': 1,
            'html': 2,
            'html+pdf': 2,
            'pdf': 2,
        }
        source_stage = format_stages[self.inform]
        target_stage = format_stages[self.outform]
        if source_stage > target_stage:
            msg = (
                "Conversion direction must be jats -> (json|yaml|jsoml) -> (html|html+pdf|pdf)"
            )
            self.parser.error(msg)

    def just_copy(self):
        if self.inform == self.outform:
            if self.inform not in ["json", "yaml", "jsoml"]:
                if self.inpath.is_dir():
                    copytree_nostat(self.inpath, self.outpath)
                else:
                    shutil.copy(self.inpath, self.outpath)
                return True
        return False

    def check_imports(self, import_names, act):
        form = self.inform if act == "read" else self.outform
        for name in import_names:
            try:
                importlib.import_module(name)
            except ImportError as e:
                msg = f"{e.name} must be installed to {act} {form}"
                self.parser.error(msg)

    def load_webstract(self):
        if self.inform == "jats":
            self.check_imports(["elifetools", "jsoml"], "read")
            from epijats.jats import webstract_from_jats
            return webstract_from_jats(self.inpath, self.config.pandoc_opts)
        elif self.inform == "json":
            return Webstract.load_json(self.inpath)
        elif self.inform == "yaml":
            self.check_imports(["ruamel.yaml"], "read")
            return Webstract.load_yaml(self.inpath)
        elif self.inform == "jsoml":
            self.check_imports(["jsoml"], "read")
            return Webstract.load_xml(self.inpath)
        return None

    def convert(self, webstract):
        if self.outform == "json":
            webstract.dump_json(self.outpath)
        elif self.outform == "yaml":
            self.check_imports(["ruamel.yaml"], "write")
            webstract.dump_yaml(self.outpath)
        elif self.outform == "jsoml":
            self.check_imports(["jsoml"], "write")
            webstract.dump_xml(self.outpath)
        else:
            assert self.outform in ["html", "html+pdf", "pdf"]
            self.check_imports(["jinja2"], "write")
            with tempfile.TemporaryDirectory() as tempdir:
                if self.outform == "html+pdf":
                    self.config.show_pdf_icon = True
                eprint = Eprint(webstract, Path(tempdir) / "html", self.config)
                if self.outform == "html":
                    eprint.make_html_dir(self.outpath)
                else:
                    self.check_imports(["weasyprint"], "write")
                    enable_weasyprint_logging()
                    if self.outform == "html+pdf":
                        eprint.make_html_and_pdf(
                            self.outpath,
                            self.outpath / "article.pdf"
                        )
                    elif self.outform == "pdf":
                        eprint.make_pdf(self.outpath)


def main(args=None):
    return Main(args).run()

if __name__ == "__main__":
    exit(main())



from .util import copytree_nostat
from .jinja import PackagePageGenerator


#std library
import os, tempfile
from datetime import datetime, date, time, timezone
from pathlib import Path
from pkg_resources import resource_filename


# WeasyPrint will inject absolute local file paths into a PDF file if the input HTML
# file has relative URLs in anchor hrefs.
# This hardcoded meaningless HACK_WEASY_PATH is to ensure these local file paths are
# meaningless and constant (across similar operating systems).
HACK_WEASY_PATH = Path(tempfile.gettempdir()) / "mZ3iBmnGae1f4Wcgt2QstZn9VYx"


class EprinterConfig:
    def __init__(self, theme_dir=None, dsi_base_url=None, math_css_url=None):
        self.urls = dict(
            dsi_base_url=(dsi_base_url.rstrip("/") if dsi_base_url else None),
            math_css_url=(math_css_url or "static/katex/katex.css"),
        )
        self.pandoc_opts = []
        if theme_dir:
            self.pandoc_opts = ["--data-dir", theme_dir, "--defaults", "pandoc.yaml"]
        self.article_style = 'lyon'
        self.embed_web_fonts = True
        self.show_pdf_icon = False


class Eprint:

    _gen = None
    _static_dir = Path(resource_filename(__name__, "static/"))

    def __init__(self, webstract, tmp, config=None):
        if config is None:
            config = EprinterConfig()
        self._tmp = Path(tmp)
        self._html_ctx = dict(config.urls)
        self._html_ctx["article_style"] = config.article_style
        self._html_ctx["embed_web_fonts"] = config.embed_web_fonts
        self._html_ctx["show_pdf_icon"] = config.show_pdf_icon
        self.webstract = webstract
        if Eprint._gen is None:
            Eprint._gen = PackagePageGenerator()

    def _get_html(self):
        html_dir = self._tmp
        os.makedirs(html_dir, exist_ok=True)
        ret = html_dir / "index.html"
        # for now just assume math is always needed
        ctx = dict(doc=self.webstract.facade, has_math=True, **self._html_ctx)
        self._gen.render_file("article.html.jinja", ret, ctx)
        if not ret.with_name("static").exists():
            os.symlink(Eprint._static_dir, ret.with_name("static"))
        if self.webstract.source.subpath_exists("pass"):
            self.webstract.source.symlink_subpath(html_dir / "pass", "pass")
        return ret

    def make_html_dir(self, target):
        copytree_nostat(self._get_html().parent, target)

    @staticmethod
    def copy_static_dir(target):
        copytree_nostat(Eprint._static_dir, target)

    @staticmethod
    def html_to_pdf(source, target):
        import weasyprint

        weasyprint.HTML(source).write_pdf(target)

    @staticmethod
    def stable_html_to_pdf(html_path, target, source_date):
        target = Path(target)
        os.environ.update(source_date)
        if os.environ.get("EPIJATS_SKIP_PDF"):
            return
        try:
            os.remove(HACK_WEASY_PATH)
        except FileNotFoundError:
            pass
        os.symlink(html_path.parent.resolve(), HACK_WEASY_PATH)
        Eprint.html_to_pdf(HACK_WEASY_PATH / html_path.name, target)
        os.remove(HACK_WEASY_PATH)
        return target

    def make_pdf(self, target):
        Eprint.stable_html_to_pdf(self._get_html(), target, self._source_date_epoch())

    def make_html_and_pdf(self, html_target, pdf_target):
        html_path = self._get_html()
        copytree_nostat(html_path.parent, html_target)
        Eprint.stable_html_to_pdf(html_path, pdf_target, self._source_date_epoch())

    def _source_date_epoch(self):
        ret = dict()
        if self.webstract.date is not None:
            assert isinstance(self.webstract.date, date)
            doc_date = datetime.combine(self.webstract.date, time(0), timezone.utc)
            source_mtime = doc_date.timestamp()
            if source_mtime:
                ret["SOURCE_DATE_EPOCH"] = "{:.0f}".format(source_mtime)
        return ret



from .util import swhid_from_files

import copy, json, os
from pathlib import Path
from datetime import date
from warnings import warn


SWHID_SCHEME_LENGTH = len("shw:1:abc:")


class Source:
    def __init__(self, swhid=None, path=None):
        self._swhid = swhid
        self.path = None if path is None else Path(path)
        if self._swhid is None and self.path is None:
            raise ValueError("SWHID or path must be specified")

    @property
    def swhid(self):
        if self._swhid is None:
            self._swhid = swhid_from_files(self.path)
            if not self._swhid.startswith("swh:1:"):
                raise ValueError("Source not identified by SWHID v1")
        return self._swhid

    @property
    def hash_scheme(self):
        return self.swhid[:SWHID_SCHEME_LENGTH]

    @property
    def hexhash(self):
        return self.swhid[SWHID_SCHEME_LENGTH:]

    def __str__(self):
        return self.swhid

    def __repr__(self):
        return str(dict(swhid=self._swhid, path=self.path))

    def __eq__(self, other):
        if isinstance(other, Source):
            return self.swhid == other.swhid
        return False

    def subpath_exists(self, subpath="."):
        return self.path is not None and (self.path / subpath).exists()

    def symlink_subpath(self, symlink, subpath="."):
        if self.path is None:
            raise ValueError(f"Path unknown for: {self._swhid}")
        srcpath = self.path / subpath
        if not srcpath.exists():
            raise ValueError(f"Path not found: {srcpath}")
        if symlink.exists():
            os.unlink(symlink)
        os.symlink(srcpath.resolve(), symlink)


class Webstract(dict):
    KEYS = [
        "abstract",
        "archive_date",
        "body",
        "contributors",
        "date",
        "edition",
        "source",
        "title",
    ]

    def __init__(self, init=None):
        super().__init__()
        self["contributors"] = list()
        if init is None:
            init = dict()
        for key, value in init.items():
            self[key] = value
        self._facade = WebstractFacade(self)

    @property
    def facade(self):
        return self._facade

    @property
    def source(self):
        return self["source"]

    @property
    def date(self):
        return self.get("date")

    def __setitem__(self, key, value):
        if key not in self.KEYS:
            raise KeyError(f"Invalid Webstract key: {key}")
        if value is None:
            warn(f"Skip set of None for webstract key '{key}'", RuntimeWarning)
            return
        elif key == "source":
            if isinstance(value, Path):
                value = Source(path=value)
            elif not isinstance(value, Source):
                value = Source(value)
        elif key == "date" and not isinstance(value, date):
            value = date.fromisoformat(value)
        elif key == "archive_date" and not isinstance(value, date):
            value = date.fromisoformat(value)
        super().__setitem__(key, value)

    def dump_json(self, path):
        """Write JSON to path."""

        with open(path, "w") as file:
            json.dump(
                self,
                file,
                indent=4,
                default=str,
                ensure_ascii=False,
                sort_keys = True,
            )
            file.write("\n")

    @staticmethod
    def load_json(path):
        with open(path) as f:
            return Webstract(json.load(f))

    def dump_yaml(self, path):
        from ruamel.yaml import YAML

        if isinstance(path, str):
            path = Path(path)
        if not isinstance(path, Path):
            raise NotImplementedError
        yaml = YAML()
        with open(path, "w") as file:
            plain = dict(copy.deepcopy(self))
            if "source" in plain:
                plain["source"] = str(plain["source"])
            yaml.dump(plain, file) 

    @staticmethod
    def load_yaml(source):
        from ruamel.yaml import YAML

        return Webstract(YAML(typ='safe').load(source))

    def dump_xml(self, path):
        """Write XML to path."""

        import jsoml

        with open(path, "w") as file:
            jsoml.dump(self, file)
            file.write("\n")

    @staticmethod
    def load_xml(path):
        import jsoml

        return Webstract(jsoml.load(path))


def add_webstract_key_properties(cls):
    def make_getter(key):
        return lambda self: self._webstract.get(key)

    for key in Webstract.KEYS:
        setattr(cls, key, property(make_getter(key)))
    return cls


@add_webstract_key_properties
class WebstractFacade:
    def __init__(self, webstract):
        self._webstract = webstract

    @property
    def _edidata(self):
        return self._webstract.get('edition', dict())

    @property
    def authors(self):
        ret = []
        for c in self.contributors:
            ret.append(c["given-names"] + " " + c["surname"])
        return ret

    @property
    def hash_scheme(self):
        return self.source.hash_scheme

    @property
    def hexhash(self):
        return self.source.hexhash

    @property
    def obsolete(self):
        return "newer_edid" in self._edidata

    @property
    def base_dsi(self):
        return self._edidata.get("base_dsi")

    @property
    def dsi(self):
        return (self.base_dsi + "/" + self.edid) if self._edidata else None

    @property
    def edid(self):
        return self._edidata.get("edid")

    @property
    def seq_edid(self):
        from hidos.dsi import EditionId

        edid = self._edidata.get("edid")
        if not edid:
            return None
        edid = EditionId(edid)
        return str(EditionId(edid[:-1]))

    @property
    def latest_edid(self):
        if self.obsolete:
            return self._edidata.get("newer_edid")
        else:
            return self.edid



from epijats import __main__ , Webstract

import pytest
from pathlib import Path


CASES_DIR = Path(__file__).parent / "cases"


def _run(src, dest, args=""):
    if isinstance(args, str):
        args = args.split()
    args = [str(src), str(dest)] + [str(a) for a in args]
    __main__.main(args)


@pytest.mark.parametrize("form", [
    ["--to json", Webstract.load_json],
    ["--to yaml", Webstract.load_yaml],
    ["--to jsoml", Webstract.load_xml],
])
def test_jats_to(form, tmp_path):
    subcase_dir = CASES_DIR / "webstract/basic1"
    dest = tmp_path / "output_file"
    _run(subcase_dir / "input", dest, form[0])
    load_func = form[1]
    expect = Webstract.load_json(subcase_dir / "output.json")
    assert load_func(dest) == expect


INVALID_ARGS = [
    "--from html --to jats",
    "--from html --to json",
    "--from html --to yaml",
    "--from html --to jsoml",
]

@pytest.mark.parametrize("case", INVALID_ARGS)
def test_invalid_order(case):
    with pytest.raises(SystemExit):
        cmd_line_args = case.split() + ["/dev/null", "/dev/null"]
        __main__.main(cmd_line_args)



from elifetools import parseJATS
from .webstract import Webstract, Source

#std library
import sys, subprocess
from pathlib import Path
from datetime import datetime
from time import mktime
from pkg_resources import resource_filename


def run_pandoc(args, echo=True):
    cmd = ["pandoc"] + args
    if echo:
        print(" ".join([str(s) for s in cmd]))
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr)
    return proc.stdout


def pandoc_jats_to_webstract(jats_src, pandoc_opts):
    args = [jats_src, "--from=jats", "-s", '--to', 'html']
    tmpl = resource_filename(__name__, "templates/webstract.pandoc")
    args += ["--template", tmpl, "--citeproc", "--filter=pandoc-katex-filter"]
    args += ["--metadata", "reference-section-title=References"]
    args += ["--metadata", "link-citations=true"]
    args += ["--shift-heading-level-by=1", "--wrap=preserve"]
    return run_pandoc(args + pandoc_opts)


def webstract_from_jats(src, pandoc_opts):
    src = Path(src)
    jats_src = src / "article.xml" if src.is_dir() else src
    pipeout = pandoc_jats_to_webstract(jats_src, pandoc_opts)
    ret = Webstract.load_xml(pipeout)

    ret['source'] = Source(path=src)

    soup = parseJATS.parse_document(jats_src)

    dates = parseJATS.pub_dates(soup)
    if dates:
        date = datetime.fromtimestamp(mktime(dates[0]["date"])).date()
    else:
        date = None
    ret['date'] = date

    ret['contributors'] = parseJATS.contributors(soup)
    for c in ret['contributors']:
        if 'orcid' in c:
            c['orcid'] = c['orcid'].rsplit("/", 1)[-1]

    return ret



import jinja2


class SuccessionFacade:
    def __init__(self, succession):
        self.succession = succession

    @property
    def dsi(self):
        return self.succession.dsi

    @property
    def ref_commit(self):
        return self.succession.ref_commit.hexsha

    @property
    def sign_key(self):
        fingerprint = self.succession.sign_key_fingerprint
        return fingerprint.hex().upper()


class WebPageGenerator:
    def __init__(self):
        self.env = jinja2.Environment(
            loader=jinja2.ChoiceLoader([]),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
            extensions=["jinja2.ext.do"],
        )

    def add_template_loader(self, loader):
        self.env.loader.loaders.append(loader)

    def render_file(self, tmpl_subpath, dest_filepath, ctx=dict()):
        tmpl = self.env.get_template(str(tmpl_subpath))
        tmpl.stream(**ctx).dump(str(dest_filepath), "utf-8")


class PackagePageGenerator(WebPageGenerator):
    def __init__(self):
        super().__init__()
        self.add_template_loader(jinja2.PackageLoader(__name__, "templates"))


def style_template_loader():
    return jinja2.PrefixLoader(
        {"epijats": jinja2.PackageLoader(__name__, "templates/epijats")}
    )



import pytest

import tempfile
from os import listdir
from pathlib import Path

import hidos

from epijats import util, Webstract, EprinterConfig, DocLoader
from epijats.jats import webstract_from_jats


CONFIG = EprinterConfig()

CASES_DIR = Path(__file__).parent / "cases"

WEBSTRACT_CASES = [f"webstract/{s}" for s in listdir(CASES_DIR / "webstract")]
SUCCESSION_CASES = listdir(CASES_DIR / "succession")

EDITION_CASES = list()
for s in SUCCESSION_CASES:
    for e in listdir(CASES_DIR / "succession" / s):
        EDITION_CASES.append(f"succession/{s}/{e}")

ARCHIVE = hidos.Archive(".", unsigned_ok=True)


@pytest.mark.parametrize("case", WEBSTRACT_CASES)
def test_webstracts(case):
    got = webstract_from_jats(CASES_DIR / case / "input", CONFIG.pandoc_opts)
    expect = Webstract.load_json(CASES_DIR / case / "output.json")
    assert got == expect


@pytest.mark.parametrize("case", SUCCESSION_CASES)
def test_editions(case):
    with tempfile.TemporaryDirectory() as tmpdir:
        loader = DocLoader(tmpdir, CONFIG)
        succ = ARCHIVE.find_succession(case)
        for edition in succ.root.all_subeditions():
            if edition.has_digital_object:
                got = loader.webstract_from_edition(edition)
                edition_path = CASES_DIR / "succession" / case / str(edition.edid)
                expect = Webstract.load_json(edition_path / "output.json")
                assert got == expect


@pytest.mark.parametrize("case", WEBSTRACT_CASES + EDITION_CASES)
def test_xml(case):
    got = Webstract.load_xml(CASES_DIR / case / "output.xml")
    expect = Webstract.load_json(CASES_DIR / case / "output.json")
    assert got == expect


@pytest.mark.parametrize("case", WEBSTRACT_CASES + EDITION_CASES)
def test_yaml(case):
    got = Webstract.load_yaml(CASES_DIR / case / "output.yaml")
    expect = Webstract.load_json(CASES_DIR / case / "output.json")
    assert got == expect


def test_hash_file():
    got = util.swhid_from_files(CASES_DIR / "webstract/basic1/input/article.xml")
    assert got == "swh:1:cnt:2c0193c32db0f3d20f974b5f6f5e656e6898d56e"


def test_hash_dir():
    got = util.swhid_from_files(CASES_DIR / "webstract/basic1/input")
    assert got == "swh:1:dir:7a05d41c586ea4cbfa5a5e0021bc2a00ac8998ba"
